#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <signal.h>

#include "liblist.h"
#include "main.h"

unsigned sleep(unsigned seconds);

typedef struct passager
{
  int id;    /* Identifiant de la personne */
  int arret; /* Arret auquel elle souhaite descendre */
  struct passager *next, *prev;
} passager_t;

/* Liste chainee des passagers (tete et queue) pour le bus courant */
static passager_t *passagers_hd, *passagers_ta = NULL;

/* Nombre de places restantes dans le bus courant */
static int nb_places = MAX_PASSAGER;
static int temps_attente = 0;
static FILE *log;

extern int   correspondances[NB_ARRETS][NB_ARRETS];
extern char *arrets[NB_ARRETS];
extern int   logbus[BUS_NB];

static void perturbation(int num)
{
  if(num != SIGUSR1)
    {
      erreur("Probleme de signal\n");
    }
  temps_attente++;
  signal(SIGUSR1, perturbation);
  kill(getppid(), SIGUSR1);
}

static void traitantsigint(int num)
{
  if(num != SIGINT)
    {
      erreur("Probleme de signal\n");
    }
  if(log != NULL)
    fclose(log);
}

static int arret(int bus_num, int ligne, int arret, int sens)
{
  int n;
  char buffer[256];
  passager_t *p;

  if(logbus[bus_num] != 0)
    fprintf(log, "Arret a %s (%d, sens %s)\n  --> Personnes qui descendent : ", 
	    arrets[arret], arret, (sens == 0) ? "Aller" : "Retour");

  /* Construction du message  envoyer au gestionnaire */
  snprintf(buffer, sizeof(buffer), "BUS-%d-%d-%d-%d-%d;", 
	   bus_num, ligne, arret, sens, nb_places);

  p = passagers_hd;

  /* On parcourt l'ensemble des passagers du bus pour faire descendre
     ceux qui le souhaitent  l'arret courant */
  while(p != NULL)
    {
      /* Ce passager est-il arrive au bon arret ? */
      if(p->arret == arret)
	{
	  char id_s[16];
	  passager_t *t;

	  /* On ajoute l'identifiant de la personne au message 
             envoyer au gestionnaire */
	  snprintf(id_s, sizeof(id_s), "%d,", p->id);
	  strcat(buffer, id_s);
	  if(logbus[bus_num] != 0)
	    fprintf(log, "%d ", p->id);

	  /* Suppression du passager de la liste */
	  t = p->next;

	  if(p->prev)
	    p->prev->next = p->next;
	  else
	    passagers_hd = p->next;

	  if(p->next)
	    p->next->prev = p->prev;
	  else
	    passagers_ta = p->prev;

	  /* Liberation du passager */
	  free(p);
	  p = t;

	  /* Il y a une place de plus dans le bus */
	  nb_places++;
	}
      else
	p = p->next;
    }

  /* Verification subsidiaire */
  if(nb_places > MAX_PASSAGER || nb_places < 0)
    erreur("Nombre de passager impossible");

  /* On retire la derniere virgule uniquement si cela est necessaire */
  if(buffer[strlen(buffer)-1] == ',')
    buffer[strlen(buffer)-1] = '\0';

  /* On envoie l'identification du bus avec la liste des personnes qui descendent */
  //  printf("Bus[%d]-->Gestionnaire : %s\n", bus_num, buffer);
  write(fdin[bus_num][1], buffer, strlen(buffer) + 1);

  /* Attente d'une confirmation avec la liste des personnes qui montent */
  n = 0;
  do
    {
      read(fdout[bus_num][0], buffer + n, 1);
      n++;
    } while(buffer[n-1] != '\0');

  /* Verification du header du message recu */
  if(strncmp(buffer, "OK-", 3) != 0)
    erreur("Message incorrect en provenance du gestionnaire");

  if(logbus[bus_num] != 0)
    fprintf(log, "\n  --> Personnes qui montent : ");

  n = 3;
  
  /* Traitement de la liste des personnes qui montent */
  while(buffer[n] != '\0')
    {
      int j;
      char tmp[16];
      int pers_id, dest;
      passager_t *passager;
      
      /* Extraction du numero de personne */
      memset(tmp, 0, sizeof(tmp));
      for(j = 0; buffer[n] != ','; j++, n++)
	tmp[j] = buffer[n];
      pers_id = atoi(tmp);

      /* Extraction de la destination de cette personne */
      memset(tmp, 0, sizeof(tmp));
      for(j = 0, ++n; buffer[n] != ';' && buffer[n] != '\0'; j++, n++)
	tmp[j] = buffer[n];
      dest = atoi(tmp);

      /* On passe le ';' */
      if(buffer[n] != '\0')
	n++;

      /* Creation du nouveau passager */
      passager = malloc(sizeof(passager_t));
      if(passager == NULL)
	erreur("Erreur d'allocation mmoire\n");

      if(logbus[bus_num] != 0)
	fprintf(log, "%d ", pers_id);

      /* Initialisation du nouveau passager */
      passager->id = pers_id;
      passager->arret = dest;
      passager->next = passager->prev = NULL;

      /* Ajout du passager dans la liste de passagers */
      if(passagers_hd == NULL)
	passagers_hd = passagers_ta = passager;
      else
	{
	  passager->next = passagers_hd;
	  passagers_hd->prev = passager;
	  passagers_hd = passager;
	}
	  
      /* Un passager est monte --> une place de moins dans le Bus */
      nb_places--;
    }

  if(logbus[bus_num] != 0)
    fprintf(log, "\n\n");
  fflush(log);
  
  return 0;
}

/* Fonction principale du bus 
   line : numero de la ligne sur laquelle circule le bus
   bus_num : numero du bus (dans le tableau des fd)
 */
int bus(int line, int bus_num)
{
  int sens;
  int i;
  char buffer[16];
  sigset_t ens;

  srand(getpid()+bus_num);

  /* On ferme les descripteurs inutiles */
  for (i = 0; i < FD_NB; i++)
    {
      if(i != bus_num)
	{
	  close(fdin[i][1]);
	  close(fdout[i][0]);
	}
      close(fdin[i][0]);
      close(fdout[i][1]);
    }

  /* Configuration du signal SIGUSR1 */
  sigemptyset(&ens);
  signal(SIGUSR1, perturbation);
  signal(SIGINT, traitantsigint);
  sigprocmask(SIG_SETMASK, &ens, NULL);

  /* Ouverture du fichier de log si ncessaire */
  if(logbus[bus_num] != 0)
    {
      snprintf(buffer, sizeof(buffer), "bus-%d.log", bus_num);
  
      log = fopen(buffer, "w+");
      if(log == NULL)
	erreur("Impossible d'ouvrir le fichier de log");
      
      fprintf(log, "Fichier de log du bus %d circulant sur la ligne %d\n", bus_num, line);
      fflush(log);
    }

  if(bus_num > NB_LIGNES)
    sleep(RAND(10.0));

  while(1)
    {

      /* Sens aller */
      sens = FORWARD;
      for(i = 0; lignes[line][i] != -1; i++)
	{
	  arret(bus_num, line, lignes[line][i], sens);

	  if(lignes[line][i+1] != -1)
	    {
	      if (correspondances[lignes[line][i]][lignes[line][i+1]] == 0)
		erreur("Fatal error\n");
	      
	      temps_attente += correspondances[lignes[line][i]][lignes[line][i+1]];

	      while(temps_attente > 0)
		{
		  sleep(1);
		  temps_attente--;
		}
	    }
	}

      /* Sens retour */
      sens = BACKWARD;
      i -= 1;
      for( ; i >= 0; i--)
	{
	  arret(bus_num, line, lignes[line][i],  sens);

	  if(i > 0)
	    {
	      if (correspondances[lignes[line][i]][lignes[line][i-1]] == 0)
		erreur("Fatal error\n");
	      
	      temps_attente += correspondances[lignes[line][i]][lignes[line][i-1]];

	      while(temps_attente > 0)
		{
		  sleep(1);
		  temps_attente--;
		}
	    }
	}
    }


  return 0;
}
