#ifndef __LIBLIST_H__
#define __LIBLIST_H__

/*
 * Circular doubly-linked lists Implementation
 * KOS Project
 * (C) D. Decotigny 2001
 */

#define list_init(list,item) ({ \
  (item)->next = (item)->prev = (item); \
  (list) = (item); \
})

#define list_get_head(list) \
  (list)

#define list_get_tail(list) \
  ((list)?((list)->prev):NULL)

/* Caution: before_this and item are expected to be valid ! */
#define list_insert_before(before_this,item) ({ \
   (before_this)->prev->next = (item); \
   (item)->prev = (before_this)->prev; \
   (before_this)->prev = (item); \
   (item)->next = (before_this); \
})

/* Caution: after_this and item are expected to be valid ! */
#define list_insert_after(after_this,item) ({ \
   (after_this)->next->prev = (item); \
   (item)->next = (after_this)->next; \
   (after_this)->next = (item); \
   (item)->prev = (after_this); \
})

#define list_add_head(list,item) ({ \
  if (list) \
    list_insert_before(list,item); \
  else \
    list_init(list,item); \
  (list) = (item); \
})

#define list_add_tail(list,item) ({ \
  if (list) \
    list_insert_before(list,item); \
  else \
    list_init(list,item); \
})


/* Caution: no check whether item really is in list ! */
#define list_delete(list,item) ({ \
  if ( ((item)->next == (item)) && ((item)->prev == (item)) ) \
    (item)->next = (item)->prev = (list) = NULL; \
  else { \
    (item)->prev->next = (item)->next; \
    (item)->next->prev = (item)->prev; \
    if ((item) == (list)) (list) = (item)->next; \
    (item)->prev = (item)->next = NULL; \
  } \
})  

#define list_foreach_forward(list,iterator,nb_elements) \
        for (nb_elements=0, (iterator) = (list) ; \
             (iterator) && (!nb_elements || ((iterator) != (list))) ; \
             nb_elements++, (iterator) = (iterator)->next )

#define list_foreach_backward(list,iterator,nb_elements) \
        for (nb_elements=0, (iterator) = list_get_tail(list) ; \
             (iterator) && (!nb_elements || ((iterator) != (list))) ; \
             nb_elements++, (iterator) = (iterator)->prev )

#define list_foreach list_foreach_forward

#endif /* __LIBLIST_H__ */
