#include <stdio.h>
#include <unistd.h>
#include <sys/types.h>
#include <unistd.h>
#include <stdlib.h>
#include <signal.h>
#include <ctype.h>
#include <sys/wait.h>
#include <string.h>
#include "main.h"

/* 
 * fdin : Communication vers le gestionnaire
 * fdout : Communication du gestionnaire vers les autres processus  
 */
int fdin[FD_NB][2];
int fdout[FD_NB][2];
int allpid[BUS_NB+3];

int logbus[BUS_NB];
int personnelog = -1;
int personnelogflag = 0;

void killall()
{
  int i;
  for (i = 0; i < BUS_NB+3; i++)
    kill((pid_t) allpid[i], SIGINT);
}

/* 
 * Cree tous les tubes ncessaires.
 * Ajoute les descripteurs en lecture dans l'ensemble set
 * Retourne -1 si Erreur, et le numero du plus grand
 * descripteur sinon.
 */
static int init_fd(fd_set *set)
{
  int max_fd = 0;
  int i;

  /* Creation des tubes */
  for (i = 0; i < FD_NB; i++)
    {
      if(pipe(fdin[i]) != 0)
	return -1;

      if(pipe(fdout[i]) != 0)
	return -1;

      /* On ajoute le descripteur en lecture dans l'ensemble 'set' */
      FD_SET(fdin[i][0], set);

      /* Pour le select on doit connaitre le numero du descripteur
         ouvert a scanner le plus grand */
      if(max_fd < fdin[i][0])
	max_fd = fdin[i][0];
    }

  return max_fd;
}

/*
 * Cree autant de bus que necessaire 
 */
static int bus_create()
{
  int i;
  int pid;

  for (i = 0; i < BUS_NB; i++)
    {
      switch(pid = fork())
	{
	case -1:
	  return -1;
	  break;
	case 0:
	  bus(i%5, i);
	  erreur("[Bus_create] : on ne devrait jamais arrive ici\n");
	  break;
	default:
	  allpid[i] = pid;
	}
    }

  return 0;
}

/*
 * Cree le generateur de personne 
 */
static int personne_create()
{
  int pid;

  switch(pid = fork())
    {
    case -1:
      return -1;
      break;
    case 0:
      generateur_personne();
      erreur("[Personne_create] : on ne devrait jamais arriver ici\n");
      break;
    default:
      allpid[BUS_NB+1] = pid;
    }

  return 0;
}

/*
 * Cree le gestionnaire
 */
static int gestionnaire_create(fd_set *set, int maxfd)
{
  int pid;

  switch(pid = fork())
    {
    case -1:
      return -1;
      break;
    case 0:
      gestionnaire(set, maxfd);
      erreur("[Personne_create] : on ne devrait jamais arriver ici\n");
      break;
    default:
      allpid[BUS_NB] = pid;
    }

  return 0;
}


int analyze_option(int argc, char **argv)
{
  unsigned int i;
  int j;

  while ((i = getopt (argc, argv, "b:p:")) != -1)
    {
      switch(i)
	{
	case 'b':
	  if(!strcmp(optarg, "all"))
	    {
	      for (j = 0; j < BUS_NB; j++)
		logbus[j] = 1;
	    }
	  else 
	    {
	      for (j = 0; j < strlen(optarg); j++)
		if(!isdigit((int)optarg[j]))
		  erreur("Erreur dans l'argument de bus\n");
	      if(atoi(optarg) >= BUS_NB || atoi(optarg) < 0)
		erreur("Erreur dans l'argument de bus\n");
	      
	      logbus[atoi(optarg)] = 1;
	    }
	  break;
	case 'p':
	  if(!strcmp(optarg, "none"))
	    break;
	  
	  if(!strcmp(optarg, "all"))
	    {
	      personnelogflag = 1;
	      break;
	    }

	  for (j = 0; j < strlen(optarg); j++)
	    if(!isdigit((int)optarg[j]))
	      erreur("Erreur dans l'argument de personne\n");

	  if(atoi(optarg) >= MAX_PERSONNE || atoi(optarg) < 0)
	    erreur("Erreur dans l'argument de personne\n");

	  personnelogflag = 1;
	  personnelog = atoi(optarg);
	  break;
	}
    }
  return 0;
}

int main(int argc, char **argv)
{
  int i;
  int max_fd;
  fd_set set;
  sigset_t ens;

  for (i = 0; i < BUS_NB; i++)
    logbus[i] = 0;

  analyze_option(argc, argv);

  FD_ZERO(&set);
  srand(getpid());

  reseau_init();

  max_fd = init_fd(&set);
  if(max_fd == -1)
    erreur("Erreur lors de la creation des tubes\n");

  sigfillset(&ens);
  sigprocmask(SIG_SETMASK, &ens, NULL);
  sigemptyset(&ens);

  if(bus_create() == -1)
    erreur("Erreur lors de la creation des bus\n");
 
  if(personne_create() == -1)
    erreur("Erreur lors de la creation du generateur de personne\n");

  /* On ferme les descripteurs inutiles */
  for (i = 0; i < BUS_NB; i++)
    {
      close(fdin[i][1]);
      close(fdout[i][0]);
    }

  if(gestionnaire_create(&set, max_fd) != 0)
    erreur("Impossible de creer le gestionnaire\n");

  allpid[BUS_NB+2] = getpid();
  perturbateur();

  erreur("[Main] On ne devrait jamais arriver ici\n");
  return 0;
}
