#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "liblist.h"
#include "main.h"


static int *floyd_matrix = NULL;

char *arrets[] =
{
  "Daudet",           /* Arret  0 */
  "Hugo",             /* Arret  1 */
  "Balzac",           /* Arret  2 */
  "Rousseau",         /* Arret  3 */
  "Malraux",          /* Arret  4 */
  "Rimbaud",          /* Arret  5 */
  "Ionesco",          /* Arret  6 */
  "Racine",           /* Arret  7 */
  "Molire",          /* Arret  8 */
  "Verlaine",         /* Arret  9 */
  "Shakespeare",      /* Arret 10 */
  "Chateaubriand",    /* Arret 11 */
  "Maupassant",       /* Arret 12 */
  "Voltaire",         /* Arret 13 */
  "Prvert",          /* Arret 14 */
  "Descartes",        /* Arret 15 */
  "La Fontaine",      /* Arret 16 */
  "La Bruyre",       /* Arret 17 */
  "Montesquieu",      /* Arret 18 */
  "De Musset",        /* Arret 19 */
  "Zola"              /* Arret 20 */
};

int lignes[5][7]=
{
  {  0,  1,  2,  3,  4,  5,  -1}, /* ligne 0 */
  {  6,  7,  8,  2,  9, 10,  -1}, /* ligne 1 */
  {  5, 11, 12, 13, 14,  6,  -1}, /* ligne 2 */
  { 15, 13, 16,  2, 17, 18,  -1}, /* ligne 3 */
  { 19,  0, 18, 20, 10,  4,  -1}  /* ligne 4 */
};

int correspondances[21][21]=
{
         /* 0  1  2  3  4  5  6  7  8  9 10 11 12 13 14 15 16 17 18 19 20 */
  /* 0*/  { 0, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 2, 0},
  /* 1*/  { 2, 0, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
  /* 2*/  { 0, 3, 0, 3, 0, 0, 0, 0, 2, 3, 0, 0, 0, 0, 0, 0, 4, 3, 0, 0, 0},
  /* 3*/  { 0, 0, 3, 0, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
  /* 4*/  { 0, 0, 0, 2, 0, 4, 0, 0, 0, 0, 6, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
  /* 5*/  { 0, 0, 0, 0, 4, 0, 0, 0, 0, 0, 0, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0},
  /* 6*/  { 0, 0, 0, 0, 0, 0, 0, 3, 0, 0, 0, 0, 0, 0, 4, 0, 0, 0, 0, 0, 0},
  /* 7*/  { 0, 0, 0, 0, 0, 0, 3, 0, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
  /* 8*/  { 0, 0, 2, 0, 0, 0, 0, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
  /* 9*/  { 0, 0, 3, 0, 0, 0, 0, 0, 0, 0, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
  /*10*/  { 0, 0, 0, 0, 6, 0, 0, 0, 0, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5},
  /*11*/  { 0, 0, 0, 0, 0, 4, 0, 0, 0, 0, 0, 0, 4, 0, 0, 0, 0, 0, 0, 0, 0},
  /*12*/  { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 4, 0, 3, 0, 0, 0, 0, 0, 0, 0},
  /*13*/  { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 3, 0, 3, 5, 4, 0, 0, 0, 0},
  /*14*/  { 0, 0, 0, 0, 0, 0, 4, 0, 0, 0, 0, 0, 0, 3, 0, 0, 0, 0, 0, 0, 0},
  /*15*/  { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 0, 0, 0, 0, 0, 0, 0},
  /*16*/  { 0, 0, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 4, 0, 0, 0, 0, 0, 0, 0},
  /*17*/  { 0, 0, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 0, 0},
  /*18*/  { 5, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 0, 0, 4},
  /*19*/  { 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
  /*20*/  { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 0, 0, 0, 0, 0, 0, 0, 4, 0, 0}
};

static int get_val(int *p, int i, int j, int n)
{
  return * (p + (i * n) + j);
}

static void set_val(int *p, int i, int j, int n, int val)
{
  *(p + (i * n) + j) = val;
}

/* Calcul de la matrice de Floyd pour la recherche des plus courts
   chemins */
static int *floyd(int *a)
{
  int *b, *p;
  int i, j, k;
  
  b = malloc(sizeof(correspondances));
  p = malloc(sizeof(correspondances));
  memcpy(b, a, sizeof(correspondances));
  
  for (i=0; i < NB_ARRETS; i++)
    for (j=0; j < NB_ARRETS; j++)
      set_val(p, i, j, NB_ARRETS, -1);
  
  for (k= 0; k< NB_ARRETS; ++k)
    for (i= 0; i< NB_ARRETS; ++i)
      for (j= 0; j< NB_ARRETS; ++j)
	if((get_val(b, i, k, NB_ARRETS) + get_val(b, k, j, NB_ARRETS)) 
	   < get_val(b, i, j, NB_ARRETS))
	  {
	    set_val(b, i, j, NB_ARRETS, 
		    (get_val(b, i, k, NB_ARRETS) + get_val(b, k, j, NB_ARRETS)));
	    set_val(p, i, j, NB_ARRETS, k);
	  }
  
  return p;
}


/* A partir du numero d'arret et d'un numero de ligne renvoie le
   numero de l'arret sur la ligne, ou -1 si l'arret n'existe pas sur
   la ligne */
static int cherche_arret(int arret, int ligne)
{
  int j;

  for(j = 0; lignes[ligne][j] != -1; j++)
    if(lignes[ligne][j] == arret)
      return j;

  return -1;
}

/* On cherche la ligne sur laquelle se trouvent les arrets 'deb' et
   'fin'. On retourne le numero de la ligne si elle existe et -1
   sinon. On place dans sens le sens de parcours de la ligne pour
   aller de deb  fin */

static int cherche_ligne(int deb, int fin, int *sens)
{
  int i;

  for (i = 0; i < NB_LIGNES; i++)
    {
      int first;
      int second;
      
      first = cherche_arret(deb, i);
      second = cherche_arret(fin, i);
      if(first != -1 && second != -1)
	{
	  if(first < second)
	    *sens = FORWARD;
	  else
	    *sens = BACKWARD;
	  return i;
	}

    }
  return -1;
}

/* Calcule la liste des arrets par lesquels la persone doit passer, en
   utilisant la matrice de Floyd */
static changement_t *chemin (int i, int j, changement_t *list) 
{
  int k;
  changement_t *new;

  if(floyd_matrix == NULL)
    erreur("Matrice de Floyd non initialisee\n");

  k = get_val(floyd_matrix, i, j, NB_ARRETS);

  if (k != -1) 
    {
      chemin(i,k,list);
      /* Ajout de l'arret dans le trajet de la personne */
      new = malloc(sizeof(changement_t));
      new->num = k;
      list_add_tail(list, new);
      chemin(k,j,list);
    }

  return list;
}

/* Construction de la liste des changements que doit effectuer une
   personne pour arriver  bon port */
void calculer_changements(personne_t *personne)
{
  changement_t *new;
  int n;
  int oldline;
  int newline;
  int sens;

  personne->liste_changement = NULL;

  /* On ajoute le point de depart */
  new = malloc(sizeof(changement_t));
  new->num = personne->depart;
  list_add_head(personne->liste_changement, new);

  /* On ajoute tous les arrets par lesquels la personne doit passer */
  personne->liste_changement = chemin(personne->depart, 
				      personne->arrivee, 
				      personne->liste_changement);

  /* On ajoute le point d'arrivee */
  new = malloc(sizeof(changement_t));
  new->num = personne->arrivee;
  list_add_tail(personne->liste_changement,new);

  oldline = 12;

  /* On parcourt la liste pour limner les arrets inutiles (ceux qui
     ne sont pas des correspondances) */
  list_foreach(personne->liste_changement, new, n)
    {
      newline = cherche_ligne(new->num, new->next->num, &sens);

      /* Si l'arret est sur la mme ligne que l'arret prcdent alors
         ce n'est pas une correspondance et on peut le supprimer. On
         fait bien attention  conserver le dernier arret qui est
         l'arret d'arrivee */
      if (oldline == newline && new != list_get_tail(personne->liste_changement))
	{
	  changement_t *tmp;
	  new->prev->next = new->next;
	  new->next->prev = new->prev;
	  tmp = new;
	  new = new->prev;
	  free(tmp);
	}
      else
	{
	  new->ligne = newline;
	  new->sens  = sens;
	}

      oldline = newline;
    }

  return;
}

/* Initialisation du reseau */
int reseau_init()
{
  int i, j;

  for(i = 0; i < 21; i++)
    for(j =0; j < 21; j++)
      {
	/* Pour Floyd, la diagonale doit etre  0 */
        if(i == j)
          correspondances[i][j] = 0;
	/* Toutes autres valeurs qui ne sont pas des arcs doivent tre
           l'infini (ici une valeur plus grande que toutes les autres
           valeurs) */
        if(i != j && (correspondances[i][j]) == 0)
          correspondances[i][j] = 100;
      }

  floyd_matrix = floyd((int *) correspondances);

  return 0;
}
